@@ -2,4 +2,5 @@
 The :mod:`skmine.periodic` implements tools to tackle periodicity in
 datasets, preferably presentend in the form of event logs
 """
-from .cycles import PeriodicCycleMiner
+from .cycles import PeriodicCycleMiner, SingleEventCycleMiner
+from .tree import PeriodicPatternMiner
